#!/usr/bin/env perl

use strict;
use warnings;
use diagnostics;
use Getopt::Long;
use File::Copy;
#use Pod::Text;
use File::Which qw(which); # can be removed if exonerate full path is provided for ${exonerate_bin}
use DBI;
#use Time::Format qw(%time);
use Term::ANSIColor;
use lib '/banks/_Genomics/Prog/SoS_1.x'; # to adapt with local path of SoS perl modules
use locus_at_ensembl;
use loci_from_Exonerate;
use pdb2fas;
use species;
#use PyMOL_script;
use Carp ();
local $SIG{__WARN__} = \&Carp::cluck;


#################### Parameters ####################
#local parameters
my $cache="./";                            #Where result and temporary files will be
my $exonerate_bin=which('exonerate');      #Or the full path of exonerate if not in the PATH
                                           # like my $exonerate_bin='/home/truc/bin/exonerate';
my $db='/banks/_Genomics/';                #Local repository where the chromosome files are
#If local PDB mirror
my $local_pdb='/banks/_Structures/PDB/all/pdb';  #Repository with all the PDB gzip file
#EnsEMBL db parameters
my $ensembl_release=40;                    #When databases structure doesn't change, only release nbr changes
my $host='ensembldb.ensembl.org';
my $user='anonymous';
####################################################
my $version='1.2.13';
#my $date=$time{'yyyy/mm/dd_hh:mm:ss.mmm'};
my $date=`/usr/bin/env date +%y-%m-%d_%T`;
chomp($date);


=head1 NAME

SoS.pl - SNPs mapping onto Structures

=head1 SYNOPSIS

B<SoS.pl> --pdb=PDB_ID [options]

OR     --file=local_structure_file --orgm=Genus_species --chr=n [options]

options :

=over 4

=item I<--chain>=X

=item To specify a chain (default null; the chain of the PDB file is used if only one)

=item I<--chr>=n

=item To specify the chromosome n, else the peptide locus is sought at EnsEMBL

=item I<--orgm>=Genus_species

=item The species where to seek for SNPs loci (default if 'Homo_sapiens')

=item I<--tmp>

=item To delete temporary files (default keep them all)

=item I<--noSNP>

=item To build a PyMOL file even if SoS found no SNP

=item I<--version>

=item To print current version of SoS

=item I<--help>

=item To print this help message

=back

=head1 EXAMPLES

=over 4

=item Example:

=item SoS.pl --pdb=1hcq --chain=a --tmp

=back

=head1 DESCRIPTION

SoS will :

=over 4

=item - extract the specified chain of the PDB file

=item - build the fasta sequence of this peptide

=item - find its localization in the target genome

=item - align the peptide with the genomic region

=item - extract exact matching positions

=item - check if they are variations at EnsEMBL

=item - build a PyMOL script with variations mapped on the structure

=back

=head1 REQUIREMENT

=over 4

=item B<Perl 5.6 or better> is required

=item these additionnal perl modules are required:

=item - Getopt::Long   to manage options

=item - DBI            to be able to talk with db

=item - File::Which    if you want an automatic exonerate path recognition

=item these ones must be part of your perl distribution

=item - File::Copy     to manage system copy OS independently

=item - strict

=item - warnings

=item - diagnostics

=back

=head1 VERSION

=over 8

=item version 1.2.13

=item on Jul 20th, 2006

=back

=head1 AUTHOR

=over 8

=item Sebastien MORETTI

=item sebastien.moretti@igs.cnrs-mrs.fr

=back

=cut

#Options
my ($tmp_files,$html_files,$noSNP)=(0,0,0);
my ($help,$Version)=(0,0);
my ($PDB_ID,$PDB_chain,$chr,$orgm,$file)=('','','','Homo_sapiens','');
GetOptions("pdb=s"      => \$PDB_ID,
           "chain=s"    => \$PDB_chain,
           "chr=s"      => \$chr,
           "orgm=s"     => \$orgm,
           "tmp"        => \$tmp_files,
           "file=s"     => \$file,
           "noSNP"      => \$noSNP,
           "help"       => \$help,
           "version"    => \$Version,
           "html"       => \$html_files  ######## to do ###########
	   );
#Print version information
my $doc=which('perldoc') || '';
if ($Version==1){
    print "#SoS                SNPs onto Structures
#                   Lab. Information Genomique et Structurale
#                   http://www.igs.cnrs-mrs.fr/
#                   moretti\@igs.cnrs-mrs.fr
#Version          = $version
#Current_Date     = $date
#for EnsEMBL      = v${ensembl_release}\n";
    exit;
}
#Print full help
elsif ($help==1){
#    my $parser = Pod::Text->new ();
#    print $parser->parse_from_file ($0);
    if ($doc ne ''){
        system("$doc $0");
    }
    else{
        print "\n\tperldoc command doesn't seem to be available\n\tto see the do
cumentation in pod format\n\n";
    }
    exit;
}
#Requirement for 'local_structure_file' usage
my $local_file_usage=0;
if ($file ne '' and -e "$file"){
    if ($chr eq ''){
	print "\n\tYou must provide a chromosome name,\n\tand a species if it is not the default one,\n\tto use a local structure file\n\n";
	exit;
    }
    $local_file_usage=1;
    my $substPDBname=sprintf("%05d", $$); #Must not be always the same if several SoS.pl execution with --file option at the same time !
    $substPDBname =~ s/^\d//;
    copy("$file","${substPDBname}.pdb");
    $PDB_ID=$substPDBname;
}
elsif ($file ne '' and ! -e "$file"){
    print "\n\tThe local structure file you gave is not reachable\n\n";
    exit;
}
#Print short help message if fake PDB_ID, no PDB_ID, or fake chain_ID
elsif ($PDB_ID eq '' or length($PDB_ID)!=4 or length($PDB_chain)>1){
    print "\n
\t     @@@                       @@@
\t    @   @                     @   @
\t    @             @@          @
\t     @           @  @          @
\t      @          @  @           @
\t       @         @  @            @
\t        @        @  @             @
\t    @   @         @@          @   @
\t     @@@                       @@@
\n\n\t          SNPs on Structures
\n\n";
    print "\n\tCannot get the PDB ID to seek for
\tTry \t\t\t$0 --pdb=PDB_ID [options] OR --file=local_structure_file --orgm=Genus_species --chr=n [options]
\toptions:
\t--chain=X\t\tTo specify a chain (default null\;
\t\t\t\tthe chain of the PDB file is used if only one)
\t--chr=n\t\t\tTo specify the chromosome n, else the peptide locus is sought at EnsEMBL
\t--orgm=Genus_species\tThe species where to seek for SNPs loci (default if 'Homo_sapiens')

\t--tmp\t\t\tTo delete temporary files (default keep them all)
\t--noSNP\t\t\tTo build a PyMOL file even if SoS found no SNP

\t--version\t\tTo print current version of SoS
\t--help\t\t\tTo print a full help message
\n";
    exit;
}


#####################################################
#                                                   #
#            SoS : SNPs on Structure                #
#                                                   #
#####################################################

#In the temporary directory
$date.="_$$";
$date =~ s/\:/\-/g;
my $temporary_dir=$PDB_ID.'__'.$date;
$cache=$cache.${temporary_dir}.'/';
mkdir $cache;
`mv ${PDB_ID}.pdb ${cache}/${PDB_ID}.pdb` if ($local_file_usage==1);
chdir $cache;


#Get the PDB structure file
$PDB_chain=uc($PDB_chain);
$PDB_ID=uc($PDB_ID);
my $residu_start=-800000000;
if ($local_file_usage==0){
    my $pdblow=lc($PDB_ID);
    unlink("${PDB_ID}.pdb.gz");
    if ( -e "${local_pdb}/pdb${pdblow}.ent.Z" ){
	copy("${local_pdb}/pdb${pdblow}.ent.Z","${PDB_ID}.pdb.gz");
    }
    elsif ( -e "${local_pdb}/pdb${pdblow}.noc.Z" ){
	copy("${local_pdb}/pdb${pdblow}.noc.Z","${PDB_ID}.pdb.gz");
    }
    else {
	system("wget -q -O ${PDB_ID}.pdb.gz 'http://www.rcsb.org/pdb/files/${pdblow}.pdb.gz'");
    }

    if ( -z "${PDB_ID}.pdb.gz" ){
	print "\n\tStructure doesn't seem to be available !\n\tRetry or check your PDB id\n\n";
	unlink("${PDB_ID}.pdb.gz");
	chdir "../"; rmdir $cache;
	exit;
    }
    print "\n\t\e[1m\e[37m\e[41m${PDB_ID}\e[m structure file has been downloaded ...\n";
}


#Get the fasta sequence of the PDB file
system("gunzip -f ${PDB_ID}.pdb.gz") if ($local_file_usage==0);
my $holes=&pdb2fas::fasta("${PDB_ID}.pdb","$PDB_chain");

if ($pdb2fas::failure==1){
    print "\n\tYou should have entered an invalide chain\n\tor the PDB file is corrupted\n\n";
    unlink("${PDB_ID}.fas","${PDB_ID}.pdb");
    chdir "../"; rmdir $cache;
    exit;
}
$residu_start=$pdb2fas::clefs[0];


#Protein localization
if ($chr =~ /^\d\d?$/ or $chr =~ /^[XYMDxymd]\w*$/){
    $chr =~ s/^0// if ($chr =~ /../);
    $chr=uc($chr);
    print "\tfor chromosome \e[1m\e[37m\e[41m${chr}\e[m\n";
}
else {#Get the protein locus from EnsEMBL
	  &locus_at_ensembl::request_at_EnsEMBL($orgm,$PDB_ID,$cache,${ensembl_release},$host,$user);

	  $chr=$locus_at_ensembl::chr;
	  print "\ton chromosome \e[1m\e[37m\e[41m${chr}\e[m\n";
}
$db=$db.$orgm.'.dna.chromosome.'.$chr.'.fa';


#Perform the mapping on the genome
`${exonerate_bin} --showquerygff --showtargetgff --bestn 1 --model protein2genome --percent 75 ${PDB_ID}.fas $db >${PDB_ID}.exonerate`;
if ( (-s "${PDB_ID}.exonerate") < 520 ){
    print "\n\texonerate cannot be executed correctly\n\tCheck if genomic sequences are available\n\tor if the chain you selected is the good one\n\n";
    unlink("${PDB_ID}.exonerate","${PDB_ID}.fas","${PDB_ID}.pdb");
    chdir "../"; rmdir $cache;
    exit;
}
else {
    print "\n\tProtein aligned to the genomic sequence by Exonerate ...\n";
}


#Get the SNPs which have matched. So, need to parse the exonerate result file.
my $genomic_length=0;
open(WC,"$db");
while(<WC>){
    if ($_ !~ /^>/){
        chomp;
        ${genomic_length}+=length($_);
    }
}
close WC;


my ($lociTOcheck)=&loci_from_Exonerate::parser("${PDB_ID}.exonerate",$genomic_length);
my %loci2check=%$lociTOcheck;
my @pos2check=keys(%loci2check);
print "\n\tPositions to test generated from the Exonerate output\n\tThey are sent to EnsEMBL to check variability ...\n";
print STDERR "[@pos2check]\n\t\t$chr\n";
my $ref_pos2check=\@pos2check; # A reference on our list to minimize memory and speed in execution
my %posValidated=&loci_from_Exonerate::SNPs($chr,$orgm,${ensembl_release},$host,$user,@$ref_pos2check);

if (%posValidated eq 0 and $noSNP==0){
    print "\n\tNo SNP found to match with your structure !\n\n";
    unlink("${PDB_ID}.exonerate","${PDB_ID}.fas","${PDB_ID}.pdb");
    chdir "../"; rmdir $cache;
    exit;
}
elsif (%posValidated eq 0 and $noSNP==1){
    print "\n\tNo SNP found to match with your structure !\n\n";
}
else{
    print "\n\tSNPs will be integrated to the structure file ...\n";
}


#Build the PyMOL and Rasmol scripts
my (%resiSyn,%resiNonSyn,%resiFrameShift);
sub varProcess{
    my ($resi,$changes,$y)=@_;

    foreach my $U (sort { $a <=> $b } (keys (%$holes))){
	$resi++ if ($resi >= $U);
    }


    $changes =~ s/^.+\t([\w\-\/\.\*]+)$/$1/;
    if ( length($changes)==1 ){
	$changes=$changes.$resi.$changes;
    }
    elsif ( length($changes)==3 and $changes =~ /^(\w)\/([\w\*])$/ ){
	$changes=$1.$resi.$2;
    }
    elsif ( length($changes)>3 and $changes =~ /^(\w)\/([\w\/\*]+)$/){
	$changes=$1.$resi.$2;
	$changes =~ s/\///g;
    }
    else{
	$changes="gap${resi}";
    }
    %resiSyn=(%resiSyn,$resi => $changes) if ($y =~ /\t1\t[10]\t[10]\t/);
#    %resiSyn=(%resiSyn,$resi => $changes) if ($y =~ /\t1\t0\t[10]\t/); #bug des release 33 et 34: non-syn sont en meme tps marques comme syn
    %resiNonSyn=(%resiNonSyn,$resi => $changes) if ($y =~ /\t[10]\t1\t[10]\t/);
    %resiFrameShift=(%resiFrameShift,$resi => $changes) if ($y =~ /\t[10]\t[10]\t1\t/);
}

while (my ($x,$y) = each (%posValidated)){
    print STDERR "$x\t$y\t\t\t$loci2check{$x}\n";
    if ($loci2check{$x} =~ /^\|\w\t\d+\t(\w)\t(\d+)$/){
#	print substr($subfasta,($2-1),1)."\n";;
	&varProcess($residu_start+$2-1,$y,$y) if ($1 ne 'U');
    }
    elsif ($loci2check{$x} =~ /^\|\w\t\d+\t\w\t$/){
	if ($loci_from_Exonerate::strand eq '+'){
	    my $moins=1;
	    my $debResi=0;
	    while($loci2check{($x-$moins)} !~ /^\|\w\t\d+\t\w\t\d+$/){
		$moins++;
	    }
	    my $whatAA='';
	    if ($loci2check{($x-$moins)} =~ /^\|\w\t\d+\t(\w)\t(\d+)$/){
		$whatAA=$1;
		$debResi=$2;
	    }
	    &varProcess($residu_start+$debResi-1,$y,$y) if ($whatAA ne 'U');
	}
	elsif ($loci_from_Exonerate::strand eq '-'){
	    my $moins=1;
	    my $debResi=0;
	    while($loci2check{($x+$moins)} !~ /^\|\w\t\d+\t\w\t\d+$/){
		$moins++;
	    }
	    my $whatAA='';
	    if ($loci2check{($x+$moins)} =~ /^\|\w\t\d+\t(\w)\t(\d+)$/){
		$whatAA=$1;
		$debResi=$2;
	    }
	    &varProcess($residu_start+$debResi-1,$y,$y) if ($whatAA ne 'U');
	}
    }
}

#orange=nosyn [1,0.3,0], red=nosynshift [1,0,0], green=syn [0,1,0]
my $PDB_name=${PDB_ID}.lc($PDB_chain);
$PDB_chain=$PDB_chain || $pdb2fas::Chain;
open(PYMOL,">${PDB_name}.p1m");
print PYMOL "embed ${PDB_ID}, pdb\n";
open(PDB,"${PDB_ID}.pdb");
print PYMOL $_ while(<PDB>);
close PDB;
print PYMOL "embed end\nload_embedded ${PDB_ID}\n";
print PYMOL "dss ${PDB_ID}\nhide all\ncreate nt, resn a+c+g+t+u+n+r+y+m+k+s+w+h+b+v+d\nshow cartoon, nt\nshow sticks, nt\ncolor slate, nt\n";
print PYMOL "create aa, ${PDB_ID} and ! resn a+c+g+t+u+n+r+y+m+k+s+w+h+b+v+d\nshow surface, aa\ncolor gray, aa\ncolor yellow, het\nshow stick, het\nhide (solvent)\n";
foreach my $u (sort { $a <=> $b } (keys (%resiSyn))){
    my $specify_chain=' and ! het';
    $specify_chain.=' and chain '.lc($PDB_chain) if ($PDB_chain ne '');
    my $selection="$pdb2fas::seqRes{$u}__$resiSyn{$u}";
    print PYMOL "select $selection, resi $u$specify_chain\nset_color ${u}c = [0, 1, 0]\ncolor ${u}c, $selection\n";
}
foreach my $u (sort { $a <=> $b } (keys (%resiNonSyn))){
    my $specify_chain=' and ! het';
    $specify_chain.=' and chain '.lc($PDB_chain) if ($PDB_chain ne '');
    my $selection="$pdb2fas::seqRes{$u}__$resiNonSyn{$u}";
    print PYMOL "select $selection, resi $u$specify_chain\nset_color ${u}c = [1, 0.3, 0]\ncolor ${u}c, $selection\n";
}
foreach my $u (sort { $a <=> $b } (keys (%resiFrameShift))){
    my $specify_chain=' and ! het';
    $specify_chain.=' and chain '.lc($PDB_chain) if ($PDB_chain ne '');
    my $selection="$pdb2fas::seqRes{$u}__$resiFrameShift{$u}";
    print PYMOL "select $selection, resi $u$specify_chain\nset_color ${u}c = [1, 0, 0]\ncolor ${u}c, $selection\n";
}

my $name='';
$name="set_name ${PDB_ID}, ${PDB_ID}".lc($pdb2fas::Chain)."\n" if ($pdb2fas::Chain ne '');
print PYMOL "delete ${PDB_ID}\nzoom aa\ncenter aa\nset_name aa, ${PDB_ID}\n${name}select AaA, none\ndelete AaA\n";
$PDB_chain=$PDB_chain || '_';
print PYMOL "#SoS                SNPs onto Structures
#                   Lab. Information Genomique et Structurale
#                   http://www.igs.cnrs-mrs.fr/
#                   moretti\@igs.cnrs-mrs.fr
#Version          = $version
#Current_Date     = $date
#PDB identifier   = ${PDB_ID}
#chain identifier = ${PDB_chain}
#organism         = $orgm
#on chromosome    = $chr
#from EnsEMBL     = v${ensembl_release}\n" if ($local_file_usage==0);
print PYMOL "#SoS                SNPs onto Structures
#                   Lab. Information Genomique et Structurale
#                   http://www.igs.cnrs-mrs.fr/
#                   moretti\@igs.cnrs-mrs.fr
#Version          = $version
#Current_Date     = $date
#PDB identifier   = ${file}
#chain identifier = ${PDB_chain}
#organism         = $orgm
#on chromosome    = $chr
#from EnsEMBL     = v${ensembl_release}\n" if ($local_file_usage==1);
close PYMOL;


#Remove temporary files if wanted
`mv ${PDB_name}.p1m ../` if ($local_file_usage==0);
`mv ${PDB_name}.p1m ../${file}.p1m` if ($local_file_usage==1);
unlink("${PDB_ID}.fas","${PDB_ID}.exonerate","${PDB_ID}.pdb") if ($tmp_files==1);
chdir '../';
rmdir $cache if ($tmp_files==1);


exit;
